"""Model base class."""
import functools
import inspect
from copy import deepcopy

from pydent.marshaller.descriptors import DataAccessor
from pydent.marshaller.exceptions import SchemaException
from pydent.marshaller.exceptions import SchemaModelException
from pydent.marshaller.fields import Callback
from pydent.marshaller.registry import ModelRegistry
from pydent.marshaller.schema import DynamicSchema
from pydent.marshaller.schema import SchemaRegistry


def add_schema(cls):
    """Decorator that dynamically attaches a schema to a model."""

    if not issubclass(cls, SchemaModel):
        raise SchemaException("Model must be a subclaass of '{}'".format(SchemaModel))
    new_schema = SchemaRegistry(
        SchemaRegistry.make_schema_name(cls.__name__),
        (DynamicSchema,),
        {"_model_class": cls},
    )
    new_schema.register(cls)
    return cls


class SchemaModel(metaclass=ModelRegistry):
    """A Schema class that holds information on
    serialization/deserialization."""

    __slots__ = [ModelRegistry._data_key, ModelRegistry._deserialized_key, "__dict__"]

    def __init__(self, data=None):
        """The model initializer.

        :param data: data to add to the model
        :type data: dict
        """
        if data is None:
            data = {}
        if not getattr(self, ModelRegistry._data_key, None):
            setattr(self, ModelRegistry._data_key, data)
        if not getattr(self, ModelRegistry._deserialized_key, None):
            setattr(self, ModelRegistry._deserialized_key, {})
        self.add_data(data)

    @property
    def model_schema(self):
        """Returns the attached model schema.

        :return: the model's schema class
        :rtype: DynamicSchema
        """
        return getattr(self.__class__, "model_schema", None)

    def add_data(self, data):
        """Initializes fake attributes that correspond to data."""
        if not self.model_schema:
            filepath = ""
            try:
                filepath = inspect.getfile(self.__class__)
            except Exception:
                pass
            raise SchemaModelException(
                "Cannot initialize a {} without a {}. "
                "Use '@{}' to decorate the class definition for '{}' located in {}".format(
                    SchemaModel.__name__,
                    DynamicSchema.__name__,
                    add_schema.__name__,
                    self.__class__.__name__,
                    filepath,
                )
            )
        if data is not None:
            self.__class__.model_schema.init_data_accessors(self, data)

    def _get_data(self):
        """Return the model's data."""
        return getattr(self, self.__class__._data_key)

    def _get_deserialized_data(self):
        """Return the deserialized model data."""
        return getattr(self, self.__class__._deserialized_key)

    def get_deserialized(self, name):
        """Get deserialized data by name.

        .. deprecated:: 0.1.5a7
            method will be removed in 0.2.
            Use self._get_deserialized_data[name] instead.

        :param name: name of attribute
        :return:
        """
        return self._get_deserialized_data()[name]

    def is_deserialized(self, name: str) -> bool:
        """Check if a given key has already been deserialized.

        .. versionadded:: 0.1.5a7
            Method added

        :param name: name of the attribute
        :return: True if key has been deserialized.
        """
        data = self._get_deserialized_data()
        accessor = self.fields[name]
        if name in data and data[name] is not accessor.ACCESSOR.HOLDER:
            return True
        return False

    def reset_field(self, name: str):
        """Reset the field descriptor to its placeholder value, returning the
        behavior.

        .. versionadded:: 0.1.5a7
            Method added to reset deserialized attributes

        :param name: name of the attribute
        :return: None
        """
        del self._get_deserialized_data()[name]

    @classmethod
    def _set_data(cls, data, calling_obj=None):
        instance = SchemaModel.__new__(cls)
        SchemaModel.__init__(instance, data)
        return instance

    @staticmethod
    def _keys_to_dict(keys):
        if keys is None:
            return {}
        elif isinstance(keys, str):
            return {keys: None}
        elif isinstance(keys, dict):
            return dict(keys)
        else:
            return {k: None for k in keys}

    @classmethod
    def _dump(cls, obj, only=None, include=None, ignore=None):
        if not issubclass(type(obj), SchemaModel):
            return obj

        only = cls._keys_to_dict(only)
        include = cls._keys_to_dict(include)
        ignore = cls._keys_to_dict(ignore)

        data = {}

        serialized_data = obj._get_data()

        include_and_only = set(include).union(set(only))

        for fname in include_and_only:
            getattr(obj, fname)

        keys = set(serialized_data.keys())
        keys = keys.difference(set(ignore))
        if only:
            keys = keys.intersection(set(only))

        model_fields = obj.__class__.model_schema.fields

        callback_fields = []
        always_dump = []
        for fname, field in model_fields.items():
            if hasattr(field, "nested") or issubclass(type(field), Callback):
                callback_fields.append(fname)
            if getattr(field, "always_dump", None):
                always_dump.append(fname)

        include_and_only = include_and_only.union(set(always_dump))
        callback_keys = include_and_only.intersection(set(callback_fields))
        callback_keys = callback_keys.difference(set(ignore))

        data_keys = keys.difference(set(callback_fields))

        for key in data_keys:
            data[key] = serialized_data[key]
        for key in callback_keys:
            field = model_fields[key]
            val = getattr(obj, key)
            dump = functools.partial(
                cls._dump,
                include=include.pop(key, None),
                only=only.pop(key, None),
                ignore=ignore.pop(key, None),
            )
            if getattr(model_fields[key], "nested", None) and model_fields[key].many:
                if isinstance(val, list):
                    data[field.data_key] = [dump(v) for v in val]
            else:
                data[field.data_key] = dump(val)
        return data

    def dump(self, only=None, include=None, ignore=None):
        """Dump/serializes the model to a json-like dictionary.

        :param only: restricts dump/serialization to the provided keys
        :type only: basestring|list|tuple|dict
        :param include: include any callback fields in the dump/serialization
        :type include: basestring|list|tuple|dict
        :param ignore: ignores fields in the dump/serialization
        :type ignore: basestring|list|tuple|dict
        :return: the serialized data
        :rtype: dict
        """
        return deepcopy(self._dump(self, only=only, include=include, ignore=ignore))

    # def __dir__(self):
    #     return ['a']

    @classmethod
    def load(cls, data):
        """Deserialize the data to a model.

        :param data: data to deserialize
        :type data: dict
        :return: the deserialized model
        :rtype: SchemaModel
        """
        return cls._set_data(data)

    def __setstate__(self, state):
        """Overrides how models are unpickled.

        The default way pickling works by dumping the __dict__ along
        with the class definition. Since the way data is handle in the
        marshaller object involves dynamic creation of descriptors,
        these are not properly handled upon load. Here we will create
        these descriptors upon unpickling. Unlike serialized data,
        deserialized data is handled explicitly using 'fields' in the
        schema/model definition, it is not necessary to do anything with
        these descriptors.
        """
        self.__dict__ = state

        # add data accessors
        data = state[ModelRegistry._data_key]
        for k in data:
            if k not in self.__class__.__dict__:
                setattr(self.__class__, k, DataAccessor(k, ModelRegistry._data_key))



"""Model and schema registry meta-classes."""
from collections import defaultdict
from pprint import pformat

from pydent.marshaller.exceptions import ModelRegistryError
from pydent.marshaller.exceptions import SchemaRegistryError
from pydent.utils.loggable import condense_long_lists
from pydent.utils.logging_helpers import did_you_mean


class SchemaRegistry(type):
    """Stores a list of models that can be accessed by name."""

    schemas = {}  # the registry of Schemas instantiated
    _fields = None  # list of fields instantiated
    _model_class = None  # reference to model class the schema is attached to
    BASE = "DynamicSchema"

    def __init__(cls, name, bases, selfdict):
        """Saves model to the registry."""
        super().__init__(name, bases, selfdict)
        if bases != () and name != SchemaRegistry.BASE:
            if name in SchemaRegistry.schemas:
                raise SchemaRegistryError(
                    "Cannot register '{}' since it already exists.".format(name)
                )
            SchemaRegistry.schemas[name] = cls

    @property
    def model_class(cls):
        """The registered model class."""
        return cls._model_class

    @property
    def fields(cls):
        """The shcmea fields."""
        return cls._fields

    @property
    def grouped_fields(cls):
        """Fields grouped by their base classes.

        Returns empty list of base class not in fields.
        """
        grouped = defaultdict(dict)
        for fname, field in cls.fields.items():
            mro = field.__class__.__mro__
            for b in mro[:-2]:
                grouped[b.__name__][fname] = field
        return grouped

    @staticmethod
    def make_schema_name(name):
        """Makes a Schema class name from a model name (appending 'Schema' to
        name)"""
        return name + "Schema"

    @staticmethod
    def get_schema(name):
        """Gets model by model_name."""
        if name not in SchemaRegistry.schemas:
            raise SchemaRegistryError(
                'Schema "{}" not found in SchemaRegistry. Available schemas:\n{}'.format(
                    name, ",".join(SchemaRegistry.schemas.keys())
                )
            )
        else:
            return SchemaRegistry.schemas[name]

    @staticmethod
    def get_model_schema(model_name):
        """Gets a model schema from a model name."""
        return SchemaRegistry.get_schema(SchemaRegistry.make_schema_name(model_name))


class ModelRegistry(type):
    """Stores list of all models instantiated from the SchemaModel base."""

    models = {}  # the dictionary of model classes instantiated
    _model_schema = None  # the class attribute to store the model_classes Schema class
    _fields_key = "fields"  # the class level attribute key to instantiate model fields
    _data_key = "__serialized_data"  # the attribute key used to store serialized data
    _deserialized_key = (
        "__deserialized_data"
    )  # the attribute key used to store serialized data
    BASE = "SchemaModel"

    def __init__(cls, name, bases, selfdict):
        """Saves model to the registry."""
        super().__init__(name, bases, selfdict)
        if bases != () and bases[0].__name__ != ModelRegistry.BASE:
            ModelRegistry.models[name] = cls

    @property
    def model_schema(cls):
        """Returns the class's schema."""
        return cls._model_schema

    @classmethod
    def did_you_mean_model(cls, model_name, fallback=True):
        model_names = list(cls.models.keys())
        available_models = ", ".join(model_names)
        msg = did_you_mean(model_name, model_names)
        if not msg and fallback:
            return "Available models: {}".format(pformat(available_models))
        return msg

    @classmethod
    def get_model(cls, name):
        """Gets a model class by name."""
        if name not in ModelRegistry.models:
            raise ModelRegistryError(
                "Model '{}' not found in registry.\n{}".format(
                    name, cls.did_you_mean_model(name)
                )
            )
        return cls.models[name]



"""aql.py.

Aquarium Query Language

Aquarium Query Language (:mod:`pdyent.aql`)
=============================

Validated JSON Schemas (https://json-schema.org/)

Check out the :ref:`JSON Schema page <json_schema>` for more information.
"""
import json
from copy import deepcopy
from datetime import datetime
from datetime import timedelta
from os.path import abspath
from os.path import dirname
from os.path import join
from typing import Dict
from typing import List
from typing import Union

import jsonschema

from pydent.base import ModelBase
from pydent.exceptions import AquariumQueryLanguageValidationError
from pydent.relationships import HasMany
from pydent.relationships import HasOne
from pydent.sessionabc import SessionABC

aql_schema_filepath = join(abspath(dirname(__file__)), "aql.schema.json")
with open(aql_schema_filepath, "r") as f:
    aql_schema = json.load(f)


class QueryBuilder:
    AND = " AND "
    OR = " OR "

    class Op:
        op = "="

        def __init__(self, v):
            self.v = v

    class Eq(Op):
        op = "="

    class Not(Op):
        op = "!="

    class Lt(Op):
        op = "<"

    class Gt(Op):
        op = ">"

    class Gte(Op):
        op = ">="

    class Lte(Op):
        op = "<="

    @classmethod
    def _parse_key_val(cls, k, v):
        if issubclass(v.__class__, cls.Op):
            return '{} {} "{}"'.format(k, v.op, v.v)
        else:
            return '{} {} "{}"'.format(k, cls.Eq.op, v)

    @classmethod
    def sql(cls, data):
        rows = []
        for k, v in data.items():
            if isinstance(v, list) or isinstance(v, tuple) or isinstance(v, set):
                parts = [cls._parse_key_val(k, _v) for _v in v]
                rows.append("( {} )".format(cls.OR.join(parts)))
            else:
                rows.append(cls._parse_key_val(k, v))
        return cls.AND.join(rows)


def validate_with_schema(
    data: Dict, schema: Dict, do_raise: bool = True, reraise_as: Exception = None
):
    """Validate JSON data using a JSON Schema."""
    try:
        jsonschema.validate(data, schema)
    except jsonschema.ValidationError as e:
        if do_raise:
            if reraise_as:
                raise reraise_as(str(e)) from e
            else:
                raise e
        else:
            return False
    return True


operators = {
    "__not__": QueryBuilder.Not,
    "__lt__": QueryBuilder.Lt,
    "__gt__": QueryBuilder.Gt,
    "__le__": QueryBuilder.Lte,
    "__ge__": QueryBuilder.Gte,
}


def _parse_time(v):
    timedelta_kwargs = {}

    for k in ["days", "seconds", "minutes", "hours", "weeks"]:
        key = "__{}__".format(k)
        if key in v:
            timedelta_kwargs[k] = v[key]

    return str(datetime.now() + timedelta(**timedelta_kwargs))


def parse_value(v):
    if isinstance(v, dict):
        if "__time__" in v:
            return _parse_time(v)
        else:
            for op in operators:
                if op in v:
                    return operators[op](parse_value(v[op]))
    else:
        return deepcopy(v)


def _aql(session, data, model=None):
    # parse query
    model_name = data.get("__model__", model)
    interface = getattr(session, model_name)
    new_query = {}
    for k in data["query"]:
        if k.startswith("__") and k.endswith("__"):
            continue
        if k in interface.model.fields:
            field = interface.model.fields[k]
            models = _aql(session, data["query"][k], model=field.nested)
            if issubclass(field.__class__, HasMany):
                new_query[field.attr] = [getattr(m, field.ref) for m in models]
            elif issubclass(field.__class__, HasOne):
                new_query[field.ref] = [getattr(m, field.attr) for m in models]
            else:
                raise ValueError("Field '{}' not supported".format(field.__class__))
        else:
            new_query[k] = parse_value(data["query"][k])

    # make options
    opts = {"reverse": True}
    if "__options__" in data["query"]:
        opts.update(deepcopy(data["query"]["__options__"]))

    # do query
    for k, v in new_query.items():
        if issubclass(type(v), QueryBuilder.Op):
            new_query = QueryBuilder.sql(new_query)
            break

    returned_models = interface.where(new_query, opts=opts)

    # return models
    if "__return__" in data["query"]:
        session.browser.get(returned_models, data["query"]["__return__"])
    if data.get("__as__", None) == "json":
        return [m._get_data() for m in returned_models]
    else:
        return returned_models


def validate_aql(data):
    validate_with_schema(
        data, aql_schema, reraise_as=AquariumQueryLanguageValidationError
    )


def aql(
    session: SessionABC, data: Dict, use_cache: bool = False
) -> Union[List[ModelBase], Dict]:
    """Perform a complex query a complex JSON query object.

    Check out the :ref:`JSON Schema page <json_schema>` for more information.

    :param session: Aquarium session instance
    :param data: data query
    :param use_cache: whether to inherit the cache from the provided session (default: False)
    :return:
    """
    validate_aql(data)
    with session.with_cache(using_models=use_cache) as sess:
        return _aql(sess, data)



"""
AqHTTP (:mod:`pydent.aqhttp`)
=============================

.. currentmodule:: pydent.aqhttp

Request class for making raw http requests to Aquarium

This module contains the AqHTTP class, which can make arbitrary post/put/get
requests to Aquarium and returns JSON data.

Generally, Trident users should be unable to make arbitrary requests using this
class.
Users should only access these methods indirectly through a ``Session`` or
``SessionInterface`` instance.
"""
import json
from typing import Dict

import requests
from retry import retry

from pydent.exceptions import ForbiddenRequestError
from pydent.exceptions import TridentJSONDataIncomplete
from pydent.exceptions import TridentLoginError
from pydent.exceptions import TridentRequestError
from pydent.exceptions import TridentTimeoutError
from pydent.utils import logger
from pydent.utils import pprint_data
from pydent.utils import url_build


LOGIN_RETRY_DELAY = 1
LOGIN_RETRY_BACKOFF = 1
LOGIN_RETRY_MAX_DELAY = 2


class AqHTTP:
    """Defines a Python to Aquarium server connection. Makes HTTP requests to
    Aquarium and returns JSON.

    This class should be obscured from Trident user so that users cannot
    make arbitrary requests to an Aquarium server and get sensitive
    information (e.g. User json that is returned contains api_key,
    password_digest, etc.) or make damaging posts. Instead, a
    SessionInterface should be the object that makes these requests.
    """

    TIMEOUT = 10

    def __init__(self, login: str, password: str, aquarium_url: str):
        """Initializes an aquarium session with login, password, and server.

        :param login: Aquarium login
        :type login: str
        :param aquarium_url: aquarium url to the server
        :type aquarium_url: str
        """
        self.login = login  #: the user login name
        self.aquarium_url = aquarium_url  #: the aquarium url
        self._requests_session = None  #: the requests session
        self.timeout = self.__class__.TIMEOUT  #: the timeout (s) for requests
        self._login(login, password)
        self.log = logger(name="AqHTTP@{}".format(aquarium_url))  #: the logger
        self._using_requests = True  #: if False, any HTTP requests will throw and error
        self.num_requests = 0  #: number of requests counter

    def on(self):
        """Turn on requests. When requests are off, this causes.

        :class:`ForbiddenRequestError <pydent.exceptions.ForbiddenRequestError>`
        to be raised if a request is made.

        :return: None
        """
        self._using_requests = True

    def off(self):
        """Turn off requests. Will cause.

        :class:`ForbiddenRequestError <pydent.exceptions.ForbiddenRequestError>`
        to be raised if a request is made.

        :return: None
        """
        self._using_requests = False

    @staticmethod
    def _format_response_info(
        response: requests.Response,
        include_text: bool = False,
        include_body: bool = True,
    ) -> str:
        if response is not None:
            if response.status_code >= 400:
                include_text = True
            info = dict(response.request.__dict__)
            info.update({"seconds": response.elapsed.total_seconds()})
            msg = "REQUEST: (t={seconds}s)  {method} {url}".format(**info)
            if include_body:
                body = info["body"]
                try:
                    my_json = body.decode("utf8").replace("'", '"')
                    body = json.loads(my_json)
                    body = pprint_data(body, max_list_len=10)
                except Exception:
                    pass
                msg = msg + "\n" + "BODY: {body}".format(body=body)
            if include_text:
                text = getattr(response, "text", "")
                try:
                    text = json.dumps(json.loads(text), indent=2)
                except Exception:
                    pass
                msg = msg + "\n" + "TEXT: {text}".format(text=text)
            return msg
        return "RESPONSE: NO RESPONSE"

    @staticmethod
    def _format_request_status(response: requests.Response) -> str:
        if response is not None:
            return "STATUS:  {} {}".format(response.status_code, response.reason)
        return "STATUS: NO RESPONSE"

    @property
    def url(self) -> str:
        """An alias of aquarium_url."""
        return self.aquarium_url

    @staticmethod
    def create_session_json(login: str, password: str) -> Dict:
        return {"session": {"login": login, "password": password}}

    @retry(
        TridentLoginError,
        delay=LOGIN_RETRY_DELAY,
        backoff=LOGIN_RETRY_BACKOFF,
        max_delay=LOGIN_RETRY_MAX_DELAY,
        tries=3,
    )
    def _login(self, login: str, password: str):
        """Login to aquarium and saves header as a requests.Session()

        :param login: Aquarium login
        :param password: Aquarium password
        :return: None
        :raises:
            TridentLoginError: If Aquarium authentication fails or server could not
            be contacted
            TridentTimeoutError: If response time exceeds specified timeout
        """
        session_data = self.create_session_json(login, password)
        try:
            res = requests.post(
                url_build(self.aquarium_url, "sessions.json"),
                json=session_data,
                timeout=self.timeout,
            )
            if res is None:
                raise requests.exceptions.ConnectionError
            cookies = dict(res.cookies)

            # Check for remember token
            if not any(["remember_token" in k for k in dict(res.cookies)]):
                raise TridentLoginError(
                    "Authentication error - Cannot connect to Aquarium at {}. "
                    "Please check that your login, password, and url are correct."
                )

            # fix remember token (for some outdated versions of Aquarium)
            for c in dict(cookies):
                if "remember_token" in c:
                    cookies["remember_token"] = cookies[c]
            # TODO: do we remove the session cookie to handle asynchronous requests?
            self.cookies = dict(cookies)
        except requests.exceptions.MissingSchema as error:
            raise TridentLoginError(
                "Aquarium URL {} incorrectly formatted. {}".format(
                    self.aquarium_url, error.args[0]
                )
            )
        except requests.exceptions.ConnectTimeout:
            raise TridentTimeoutError(
                "Either Aquarium took too long to respond during login or your internet"
                " connection is slow. Make sure the url {} is correct."
                " Alternatively, use Session.set_timeout"
                " to increase the request timeout.".format(self.aquarium_url)
            )
        except requests.exceptions.ConnectionError:
            raise TridentLoginError(
                "Could not contact '{}'. Login request returned no response".format(
                    self.aquarium_url
                )
            )

    @staticmethod
    def _serialize_request(url: str, method: str, body: dict) -> str:
        return json.dumps({"url": url, "method": method, "body": body}, sort_keys=True)

    @classmethod
    def _dispatch_response(cls, response):
        if response.status_code == 422:
            pass
        elif response.status_code >= 400:
            response_info = cls._format_response_info(response)
            request_status = cls._format_request_status(response)
            msg = "\n".join(
                [
                    "The Aquarium server returned an error.",
                    request_status,
                    response_info,
                ]
            )
            raise TridentRequestError(msg, response)

    def request(
        self,
        method: str,
        path: str,
        timeout: int = None,
        allow_none: bool = True,
        **kwargs,
    ) -> dict:
        """Performs a http request.

        :param method: request method (e.g. 'put', 'post', 'get', etc.)
        :type method: str
        :param path: url to perform the request
        :type path: str
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param allow_none: if False will raise error when json_data
                contains a None or null value (default: True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """

        url = url_build(self.aquarium_url, path)
        if not self._using_requests:
            raise ForbiddenRequestError(
                "Attempted a request ({} {}) when requests have been turned OFF."
                "\nDATA: {}".format(method.upper(), url, kwargs["json"])
            )

        if timeout is None:
            timeout = self.timeout
        if not allow_none and "json" in kwargs:
            self._disallow_null_in_json(kwargs["json"])

        self.num_requests += 1
        response = requests.request(
            method, url, timeout=timeout, cookies=self.cookies, **kwargs
        )

        self.log.info(self._format_response_info(response))
        self._dispatch_response(response)
        return self._response_to_json(response)

    def _response_to_json(self, response: requests.Response) -> dict:
        """Turns :class:`requests.Request` instance into a json.

        Raises TridentRequestError if an error occurs.
        """

        if response.url == url_build(self.aquarium_url, "signin"):
            msg = (
                "There was an error with authenticating the request. Aquarium "
                + "re-routed to the sign-in page."
            )
            raise TridentRequestError(msg, response)

        try:
            response_json = response.json()
        except json.JSONDecodeError:
            msg = "Response is not JSON formatted"
            msg += "\nMessage:\n" + response.text
            self.log.error(self._format_response_info(response))
            raise TridentRequestError(msg, response)
        if response_json:
            if "errors" in response_json:
                errors = response_json["errors"]
                if isinstance(errors, list):
                    errors = "\n".join(errors)
                msg = "Error response:\n{}".format(errors)
                raise TridentRequestError(msg, response)
        return response_json

    @staticmethod
    def _disallow_null_in_json(json_data: dict):
        """Raises :class:pydent.exceptions.TridentJSONDataIncomplete exception
        if json data being sent contains a null value.

        :raises:
            TridentJSONDataIncomplete: if json data containers a null or None value.
        """
        if None in json_data.values():
            raise TridentJSONDataIncomplete(
                "JSON data {} contains a null value.".format(json_data)
            )

    def post(
        self,
        path: str,
        json_data: dict = None,
        timeout: int = None,
        allow_none: bool = True,
        **kwargs,
    ) -> dict:
        """Make a post request to the session.

        :param path: url
        :type path: str
        :param json_data: json_data to post
        :type json_data: dict
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param allow_none: if False throw error if json_data contains a null
                or None value (default True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """
        return self.request(
            "post",
            path,
            json=json_data,
            timeout=timeout,
            allow_none=allow_none,
            **kwargs,
        )

    def put(
        self,
        path: str,
        json_data: dict = None,
        timeout: int = None,
        allow_none: bool = True,
        **kwargs,
    ) -> dict:
        """Make a put request to the session.

        :param path: url
        :type path: str
        :param json_data: json_data to post
        :type json_data: dict
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param allow_none: if False throw error when json_data contains a null
                or None value (default True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """
        return self.request(
            "put",
            path,
            json=json_data,
            timeout=timeout,
            allow_none=allow_none,
            **kwargs,
        )

    def get(
        self, path: str, timeout: int = None, allow_none: bool = True, **kwargs
    ) -> dict:
        """Make a get request to the session.

        :param path: url
        :type path: str
        :param timeout: time in seconds to process request before raising
                exception
        :type timeout: int
        :param allow_none: if False throw error when json_data contains a null
                or None value (default: True)
        :type allow_none: boolean
        :param kwargs: additional arguments to post to request
        :type kwargs: dict
        :return: json
        :rtype: dict
        """
        return self.request(
            "get", path, timeout=timeout, allow_none=allow_none, **kwargs
        )

    def delete(self, path: str, timeout: int = None, **kwargs) -> dict:
        return self.request("delete", path, timeout=timeout, **kwargs)

    def __repr__(self):
        return "<{}(user='{}', url='{}')>".format(
            self.__class__.__name__, self.login, self.aquarium_url
        )

    def __str__(self):
        return self.__repr__()



import json
from collections import OrderedDict
from hashlib import sha1
from typing import List

import networkx as nx

from pydent.exceptions import PlannerException
from pydent.models import Operation
from pydent.utils import Loggable
from pydent.utils import logger

# TODO: make this independent of planner
# TODO: move get_op and related methods to Plan model


class PlanOptimizationException(PlannerException):
    """Error during optimization of a plan."""


class PlanOptimizer:
    """Class that optimizes a plan."""

    def __init__(
        self, planner, inherit_logger: Loggable = None, merge_missing_samples=False
    ):
        self.planner = planner
        self.graph = self.planner.graph
        if inherit_logger:
            self.logger = inherit_logger(self)
        else:
            self.logger = logger(self)

        self.merge_missing_samples = merge_missing_samples

    # TODO: there could be a setting here to merge things with .sample=None

    def _fv_to_hash(self, fv, ft):
        # none valued Samples are never equivalent
        fvhash = OrderedDict(
            {
                "ftype": fv.field_type.ftype,
                "role": fv.role,
                "name": fv.name,
                "array": ft.array,
                "sample": OrderedDict(),
                "item": OrderedDict(),
                "value": None,
            }
        )
        if fv.sample is not None:
            sample_id = fv.child_sample_id or fv.sample.id
            fvhash["sample"]["id"] = sample_id
        elif fv.field_type.ftype != "sample":
            fvhash["value"] = fv.value
        elif fv.allowable_field_type.sample_type_id is None:
            pass
        elif not self.merge_missing_samples:
            # a deterministic hash for field values missing samples
            fvid = fv.id or fv._primary_key
            ftid = ft.id
            hashed = sha1("{}{}".format(fvid, ftid).encode("utf-8"))
            sid = hashed.hexdigest()
            fvhash["sample"] = sid

        if fv.item is not None:
            fvhash["item"]["id"] = fv.item.id
            if ft.part:
                fvhash["item"]["row"] = fv.row
                fvhash["item"]["column"] = fv.column

        return json.dumps(fvhash, indent=2)

    def _fv_array_to_hash(self, fv_array, ft, sort=True, sep="*"):
        arr = [self._fv_to_hash(fv, ft) for fv in fv_array]
        if sort:
            arr.sort()
        return sep.join(arr)

    def _op_to_hash(self, op):
        """Turns a operation into a hash using the operation_type_id, item_id,
        and sample_id."""
        ot_id = op.operation_type.id

        field_value_hashes = []
        for ft in op.operation_type.field_types:
            if ft.ftype == "sample":
                if not ft.array:
                    fv = op.field_value(ft.name, ft.role)
                    field_value_hashes.append(self._fv_to_hash(fv, ft))
                else:
                    fv_array = op.field_value_array(ft.name, ft.role)
                    field_value_hashes.append(self._fv_array_to_hash(fv_array, ft))

        field_value_hashes = sorted(field_value_hashes)
        return "{}_{}\n{}".format(
            ot_id, op.operation_type.name, "\n".join(field_value_hashes)
        )

    def _group_ops_by_hashes(self, ops):
        hashgroup = {}
        for op in ops:
            h = self._op_to_hash(op)
            hashgroup.setdefault(h, [])
            hashgroup[h].append(op)
        return hashgroup

    @staticmethod
    def _get_sample_creation_signatures(operations):
        """Return a unique signature representing how samples are being created
        by a set of operations."""
        signatures = []

        def value_hash(fv):
            item_id = None
            if not fv.child_item_id:
                if fv.item:
                    item_id = fv.item.id

            sample_id = None
            if not fv.child_sample_id:
                if fv.sample:
                    sample_id = fv.sample.id

            value = fv.value
            return "*".join(
                [
                    str(x)
                    for x in [
                        fv.name,
                        fv.role,
                        item_id,
                        sample_id,
                        value,
                        fv.row,
                        fv.column,
                    ]
                ]
            )

        for op in operations:
            signatures.append(
                "&".join(sorted([value_hash(fv) for fv in op.field_values]))
                + str(len(op.field_values))
            )
        signatures.sort()
        return tuple(set(signatures))

    def optimize(
        self,
        operations: List[Operation] = None,
        ignore: List[str] = None,
        ignore_on_the_fly: bool = True,
    ):
        """Remove redundant operations.

        :param operations: list of operations
        :param ignore: list of operation type names to ignore in optimization
        :return:
        """

        starting_signature = self._get_sample_creation_signatures(
            self.planner.operations
        )

        self.logger.info("Optimizing plan...")
        if operations is not None:
            self.logger.info(
                "   only_operations={}".format([op.id for op in operations])
            )
        self.logger.info("   ignore_types={}".format(ignore))

        # only consider operations in the 'planning state'
        if operations is None:
            operations = [
                op for op in self.planner.operations if op.status == "planning"
            ]

        # TODO: fix optimization issues
        # ignore operation type
        ignore_ops = []
        for op in operations:
            if ignore and op.operation_type in ignore:
                ignore_ops.append(op)
            elif ignore_on_the_fly and op.operation_type.on_the_fly:
                ignore_ops.append(op)

        operations = [op for op in operations if op.status == "planning"]
        nxgraph = self.graph.ops_to_subgraph(operations).nxgraph
        assert nxgraph.number_of_nodes() == len(operations)
        op_to_group_id = OrderedDict()
        group_id_to_ops = OrderedDict()

        def opkey(op):
            return op.id or op._primary_key

        # create 'group_graph'
        # group graph are all of the operations grouped into hashes
        # each group has an edge to another group
        group_graph = nx.DiGraph()

        for n in nx.topological_sort(nxgraph):
            op = nxgraph.nodes[n]["operation"]
            op_hash = self._op_to_hash(op)
            sorted_fvs = sorted(
                op.inputs, key=lambda fv: self._fv_to_hash(fv, fv.field_type)
            )
            sorted_in_wires = []
            for fv_input in sorted_fvs:
                sorted_in_wires += self.planner.get_incoming_wires(fv_input)
            sorted_src_ops = [w.source.operation for w in sorted_in_wires]
            predecessor_groups = [op_to_group_id[opkey(_op)] for _op in sorted_src_ops]
            predecessor_hash = "*".join(predecessor_groups)

            final_hash = op_hash + "__" + predecessor_hash
            op_to_group_id[opkey(op)] = final_hash
            group_id_to_ops.setdefault(final_hash, list())
            group_id_to_ops[final_hash].append(op)

            if final_hash not in group_graph:
                group_graph.add_node(final_hash, operations=[])
            group_graph.nodes[final_hash]["operations"].append(op)
            for n2 in predecessor_groups:
                group_graph.add_edge(final_hash, n2)
        # finish 'group graph'

        # do optimization
        wires_to_remove = []
        wires_to_add = set()
        ops_to_remove = []
        wires_to_add_to_array = {}

        for n1 in group_graph.nodes:
            # we look at one of the groups
            source_ops = group_graph.nodes[n1]["operations"]
            source_ops = [op for op in source_ops if op not in ignore_ops]
            # if the there is more than one operation in the group
            # perform the optimization procedure
            if len(source_ops) > 1:
                for ft in source_ops[0].operation_type.field_types:
                    if ft.role == "output":
                        source_wires = []
                        for src_op in source_ops:
                            source_wires += self.planner.get_outgoing_wires(
                                src_op.output(ft.name)
                            )
                        source_wires = set(source_wires)
                        for w in source_wires:
                            key = opkey(w.destination.operation)
                            src_fv = source_ops[0].output(w.source.name)
                            dest_group_id = op_to_group_id[key]
                            dest_op = group_id_to_ops[dest_group_id][0]
                            dest_ft = dest_op.operation_type.field_type(
                                name=w.destination.name, role="input"
                            )

                            if dest_ft.array:
                                array_key = (
                                    dest_group_id,
                                    self._fv_to_hash(w.destination, dest_ft),
                                )
                                wires_to_add_to_array.setdefault(
                                    array_key, (src_fv, dest_op, list())
                                )
                                wires_to_add_to_array[array_key][-1].append(dest_ft)
                            else:
                                dest_fv = dest_op.input(w.destination.name)
                                wires_to_add.add((src_fv, dest_fv))

        ops_to_keep = []
        for dest_ops in group_id_to_ops.values():
            dest_ops = [op for op in dest_ops if op not in ignore_ops]
            ops_to_remove += dest_ops[1:]
            ops_to_keep += dest_ops[:1]
            for op in dest_ops[1:]:
                for fv in op.field_values:
                    if fv.role == "input":
                        wires_to_remove += self.planner.get_incoming_wires(fv)
                    if fv.role == "output":
                        wires_to_remove += self.planner.get_outgoing_wires(fv)

        for w in wires_to_remove:
            self.planner.remove_wire(w.source, w.destination)

            # TODO: this optimization step may be dangerous (???)
            if w.source.operation not in ops_to_keep:
                ops_to_remove.append(w.source.operation)
            if w.destination.operation not in ops_to_keep:
                ops_to_remove.append(w.destination.operation)

        operations_list = self.planner.operations
        for op in ops_to_remove:
            if op in operations_list:
                operations_list.remove(op)

        for w in wires_to_add:
            self.planner.add_wire(*w)

        for (
            (gid, key),
            (fv_src, target_op, target_fts),
        ) in wires_to_add_to_array.items():
            key_to_inputs = {
                self._fv_to_hash(_fv, _fv.field_type): _fv for _fv in target_op.inputs
            }
            if key_to_inputs.get(key, None):
                self.planner.quick_wire(fv_src, key_to_inputs[key])
            else:
                self.planner.quick_wire(fv_src, (target_op, target_fts[0].name))

        # remove empty

        self.planner.operations = operations_list

        ending_signatures = self._get_sample_creation_signatures(
            self.planner.operations
        )
        if not starting_signature == ending_signatures:
            raise PlanOptimizationException(
                "An error occurred during optimization that would alter the "
                "intention of the Plan."
            )

        self.logger.info("\t{} operations removed".format(len(ops_to_remove)))
        self.logger.info("\t{} wires re-wired".format(len(wires_to_remove)))
        self.logger.info("\t{} new wires".format(len(wires_to_add)))

        return self.planner



from typing import Iterable
from typing import List
from typing import Union

import networkx as nx

from pydent.exceptions import TridentBaseException
from pydent.models import Collection
from pydent.models import Item
from pydent.models import Plan
from pydent.models import Sample
from pydent.sessionabc import SessionABC


VALID_INVENTORY_TYPES = [Sample, Item, Plan, Collection]
InventoryType = Union[Item, Sample, Collection, Plan]


def to_node(model):
    key = model.id or model._primary_key
    return model.__class__.__name__ + "__" + str(key)


def add_node(graph, m):
    graph.add_node(to_node(m), model=m, model_class=m.__class__.__name__)


def add_edge(graph, m1, m2):
    n1 = to_node(m1)
    if n1 not in graph:
        add_node(graph, m1)
    n2 = to_node(m2)
    if n2 not in graph:
        add_node(graph, m2)
    graph.add_edge(n1, n2)


def is_inventory_type(m):
    for valid_type in VALID_INVENTORY_TYPES:
        if isinstance(m, valid_type):
            return True
    return False


def _handle_sample(
    g: nx.DiGraph, m: InventoryType, to_visit: List[InventoryType], visited: List[str]
):
    if m.is_deserialized("field_values"):
        for fv in m.field_values:
            if fv.is_deserialized("sample"):
                add_edge(g, m, fv.sample)
                if to_node(fv.sample) not in visited:
                    to_visit.append(fv.sample)
    elif m.is_deserialized("items"):
        for item in m.items:
            if to_node(item) not in visited:
                to_visit.append(item)


def _handle_collection(g, m, to_visit):
    for pa in m.part_associations:
        if pa.has_unsaved_sample:
            add_edge(g, m, pa.part.sample)
            to_visit.append(pa.part.sample)


def _handle_item(g, m, to_visit):
    if m.is_deserialized("sample"):
        add_edge(g, m, m.sample)
        to_visit.append(m.sample)
    if m.object_type.rows:
        to_visit.append(m.containing_collection)


def _handle_plan(g, m, to_visit):
    for op in m.operations:
        for fv in op.field_values:
            if fv.sample:
                add_edge(g, m, fv.sample)
                to_visit.append(fv.sample)
            if fv.item:
                item = fv.item
                add_edge(g, m, item)
                to_visit.append(item)


def models_to_graph(session: SessionABC, models: Iterable[InventoryType]) -> nx.DiGraph:
    """Convert an iterable of inventory models (Item, Sample, Collection, or
    Plans), into a DAG.

    :param session: AqSession
    :param models: list of models
    :return: nx.DiGraph
    """

    # check inventory types
    for m in models:
        if not is_inventory_type(m):
            raise TridentBaseException(
                "{} is not a valid inventory type.".format(m.__class__.__name__)
            )

    # create graph
    graph = nx.DiGraph()
    with session.with_cache(using_models=True, timeout=60):
        for m in models:
            add_node(graph, m)

        to_visit = models[:]
        visited = []
        while to_visit:
            print(to_visit)
            m = to_visit.pop()
            if to_node(m) in visited:
                continue
            else:
                visited.append(to_node(m))
            cls = m.__class__.__name__
            if isinstance(m, Item):
                _handle_item(graph, m, to_visit)
            elif isinstance(m, Sample):
                _handle_sample(graph, m, to_visit, visited)
            elif isinstance(m, Collection):
                _handle_collection(graph, m, to_visit)
            elif isinstance(m, Plan):
                _handle_plan(graph, m, to_visit)
            else:
                raise TridentBaseException(
                    "type {} is not a valid inventory type".format(cls)
                )
    return graph


def save_inventory(
    session: SessionABC, inventory: List[InventoryType], merge_samples: bool = False
) -> List[InventoryType]:
    """Saves a list of inventory items to the server.

    :param session: the AqSession instance
    :param inventory: list of inventory items
    :param merge_samples: if True, will merge Samples by name if a sample with
        the same name already exists. See the :method:`merge_sample_by_name` method.
    :return: list of inventory that was saved.
    """
    graph = models_to_graph(session, inventory)
    new_inventory = []
    for n in nx.topological_sort(graph.reverse()):
        model = graph.nodes[n]["model"]
        if not model.id:
            new_inventory.append(model)
            if merge_samples and isinstance(model, Sample):
                model.merge()
            else:
                model.save()
    return new_inventory



"""Model serialization/deserialization schema."""
import inspect
from typing import Type

from pydent.marshaller.descriptors import DataAccessor
from pydent.marshaller.exceptions import CallbackValidationError
from pydent.marshaller.exceptions import MultipleValidationError
from pydent.marshaller.exceptions import SchemaException
from pydent.marshaller.fields import Callback
from pydent.marshaller.fields import Field
from pydent.marshaller.registry import SchemaRegistry
from pydent.marshaller.utils import make_signature_str


class DynamicSchema(metaclass=SchemaRegistry):
    """A dynamically added schema.

    Should be added using '@add_schema` decorator.
    """

    ignore = ()
    fields = dict()

    @classmethod
    def _get_model_fields(cls):
        return cls.fields

    @classmethod
    def init_field_accessors(cls):
        """Initializes callback accessors.

        :return:
        :rtype:
        """
        for field_name, field in cls._get_model_fields().items():
            field.register(field_name, cls.model_class)

    @classmethod
    def init_data_accessors(cls, instance, data, add_extra=True):
        """Initializes data accessors.

        :param instance:
        :type instance:
        :param data:
        :type data:
        :param add_extra:
        :return:
        :rtype:
        """
        if cls.model_class is not instance.__class__:
            raise SchemaException("Instance and model class are different")
        for k, v in dict(data).items():
            if k in cls.ignore:
                data.pop(k, None)
                continue
            if not add_extra and k not in cls.model_class.fields:
                raise AttributeError(
                    "Expected field missing. Cannot initialize accessor for '{}' for "
                    "'{}' because it is not a field."
                    " It may be missing from the 'field' dictionary.".format(
                        k, cls.model_class
                    )
                )
            if k not in cls.model_class.__dict__:
                setattr(cls.model_class, k, DataAccessor(k, cls.model_class._data_key))
            try:
                setattr(instance, k, v)
            except AttributeError as e:
                raise e

    @classmethod
    def validate_callbacks(cls):
        """Validates expected callback signature found in any callback in the
        registered model fields.

        :return: None
        :rtype: None
        :raises: ModelValidationError
        """
        missing_callbacks = []
        not_callable = []
        wrong_signature = []
        model_class = cls.model_class
        for rname, callback_field in cls.grouped_fields[Callback.__name__].items():
            callback_func = callback_field.callback
            if callback_func is None:
                missing_callbacks.append(
                    "Callback for {model}.{name} cannot be None".format(
                        model=model_class.__name__, name=rname
                    )
                )
                continue
            error_prefix = "Callback '{callback}' for relationship '{model}.{name}'".format(
                callback=callback_func, model=model_class.__name__, name=rname
            )
            if not callable(callback_func) and not hasattr(model_class, callback_func):
                if callback_func not in missing_callbacks:
                    missing_callbacks.append(error_prefix + " is missing.")
            else:
                args_to_send = list(callback_field.callback_args)
                kwargs_to_send = dict(callback_field.callback_kwargs)
                if isinstance(callback_func, str):
                    callback_func = getattr(model_class, callback_func)
                    args_to_send = ["self"] + args_to_send
                if not callable(callback_func):
                    not_callable.append(error_prefix + " is not callable.")
                else:
                    signature = inspect.getfullargspec(callback_func)
                    expected_args = signature.args
                    expected_kwargs = {}
                    if signature.defaults:
                        default_args = expected_args[-len(signature.defaults) :]
                        expected_args = expected_args[: -len(signature.defaults)]
                        expected_kwargs = dict(zip(default_args, signature.defaults))

                    if (
                        len(expected_args) != len(args_to_send)
                        and not signature.varargs
                    ):
                        wrong_signature.append(
                            error_prefix
                            + " expects arguments {receive_args} but would receive "
                            "arguments {sent_args}.".format(
                                receive_args=make_signature_str(
                                    expected_args, expected_kwargs
                                ),
                                sent_args=make_signature_str(
                                    list(args_to_send), kwargs_to_send
                                ),
                            )
                        )
                    else:
                        for k in kwargs_to_send:
                            invalid_keys = []
                            if k not in expected_kwargs and not signature.varkw:
                                invalid_keys.append(k)
                            if invalid_keys:
                                wrong_signature.append(
                                    error_prefix
                                    + " does not recognize named key(s) {invalid_keys} "
                                    "from signature {receive_args}.".format(
                                        invalid_keys=", ".join(
                                            ['"{}"'.format(key) for key in invalid_keys]
                                        ),
                                        receive_args=make_signature_str(
                                            expected_args, expected_kwargs
                                        ),
                                    )
                                )
                                wrong_signature.append(callback_func)

        with MultipleValidationError("") as e:
            if missing_callbacks:
                for w in missing_callbacks:
                    e.r(CallbackValidationError(w))
            if not_callable:
                for w in not_callable:
                    e.r(CallbackValidationError(w))
            if wrong_signature:
                for w in wrong_signature:
                    e.r(CallbackValidationError(w))

    @classmethod
    def register(cls, model_class: Type):
        """Registers the schema to a model class. Saves the schema class to the
        class attribute.

        :param model_class: a model class
        :type model_class: type
        :return: None
        :rtype: None
        """
        setattr(model_class, "_model_schema", cls)

        schema_fields = {}
        ignored_fields = {}
        fields_key = model_class._fields_key
        if hasattr(model_class, fields_key):
            model_fields = getattr(model_class, fields_key)

            ignore = model_fields.pop("ignore", ())
            if isinstance(ignore, str):
                ignore = (ignore,)
            cls.ignore = ignore

            for field_name, field in model_fields.items():
                if field_name in cls.ignore:
                    ignored_fields[field_name] = field
                elif issubclass(type(field), Field):
                    if field.data_key is None:
                        field.data_key = field_name
                    if issubclass(type(field), Field):
                        schema_fields[field_name] = field
            setattr(model_class, fields_key, schema_fields)
            setattr(model_class, "_ignored_fields", ignored_fields)
        setattr(cls, "_fields", schema_fields)
        setattr(cls, "_ignored_fields", ignored_fields)
        cls.init_field_accessors()
        cls.validate_callbacks()



"""
Base (:mod:`pydent.base`)
=========================

.. currentmodule:: pydent.base

Aquarium model baseclass

This module contains the base classes for Trident models.
Trident models load from JSON and dump to JSON.
This is accomplished by adding the ``@add_schema`` decorator to classes
inherited by the Base class.
Using ``@add_schema`` dynamically creates a model schema that handles dumping
and loading.

Features of Trident models:

    load - models can be loaded from JSON data. Hierarchical JSON is loaded
    intelligently.

.. code-block:: python

    Sample.load({"name": "MyPrimer", "sample_type": {"name": "Primer", ...} })
    # => <Sample(name="MyPrimer", sample_type=<SampleType(name="Primer")>)>

dump - models can be dumped to JSON. Dependent models and relationships can be
       dumped as well.

.. code-block:: python

    s.dump(include=("sample_type"))

relationships - models relationships are stored

.. code-block:: python

    s = Sample.load(
        {"name": "MyPrimer",
        "sample_type_id": 1}
    )

    primer_type = s.sample_type

"""
import itertools
from copy import deepcopy
from typing import Dict
from typing import List
from typing import Union

from inflection import tableize

from pydent.exceptions import AquariumModelError
from pydent.exceptions import NoSessionError
from pydent.exceptions import SessionAlreadySet
from pydent.interfaces import BrowserInterface
from pydent.interfaces import QueryInterface
from pydent.marshaller import fields
from pydent.marshaller import ModelRegistry
from pydent.marshaller import SchemaModel
from pydent.sessionabc import SessionABC


class ModelBase(SchemaModel):
    """Base class for Aquarium models. Subclass of.

    :class:`pydent.marshaller.MarshallerBase`

    - creates instances from JSON using `load`
    - contains a reference to the :class:`pydent.session.aqsession.AqSession`
      instance that loaded this model
    """

    PRIMARY_KEY = "id"
    GLOBAL_KEY = "rid"
    SERVER_MODEL_NAME = None
    DEFAULT_COPY_KEEP_UNANONYMOUS = ["Item", "Sample", "Collection"]
    counter = itertools.count()
    id = None
    rid = None

    def __new__(cls, *args, session=None, **kwargs):
        instance = super().__new__(cls)
        instance._session = None
        instance.session = session
        instance._rid = next(ModelBase.counter)
        return instance

    def __init__(self, **data):
        super().__init__(data)
        self.add_data({"rid": self._rid, "id": data.get("id", None)})

    @classmethod
    def _set_data(cls, data: dict, owner: "ModelBase") -> "ModelBase":
        if not hasattr(owner, "session"):
            raise NoSessionError(
                "Cannot instantiate new model because its data parent"
                " {} has no 'session' attribute".format(owner)
            )
        instance = cls.__new__(cls, session=owner.session)
        instance.raw = data
        cls.__init__(instance)
        ModelBase.__init__(instance, **data)
        return instance

    @classmethod
    def get_server_model_name(cls):
        if cls.SERVER_MODEL_NAME is None:
            return cls.__name__
        return cls.SERVER_MODEL_NAME

    @classmethod
    def get_tableized_name(cls):
        return tableize(cls.get_server_model_name())

    @property
    def rid(self) -> int:
        return self._rid

    @rid.setter
    def rid(self, _rid: int):
        self._rid = _rid
        self.add_data({"rid": self._rid})

    @property
    def _primary_key(self):
        """Returns the primary key (e.g. 'id') or the rid if id does not exist
        or is None."""
        if hasattr(self, ModelBase.PRIMARY_KEY):
            pk = getattr(self, ModelBase.PRIMARY_KEY)
            if pk:
                return pk
        return "r{}".format(self.rid)

    def append_to_many(self, name: str, model: "ModelBase") -> "ModelBase":
        """Appends a model to the many relationship.

        :param name: name of the relationship or attribute
        :type name: str
        :param model: model to append
        :type model: ModelBase
        :return: None
        :rtype: None
        """
        if name in self.get_relationships():
            field = self.get_relationships()[name]
            if not model.__class__.__name__ == field.nested:
                raise AquariumModelError(
                    "Cannot 'append_to_many.' Model must be a '{}' but found a '{}'".format(
                        field.model, model.__class__.__name__
                    )
                )
            if field.many:
                val = getattr(self, name)
                if val is None:
                    val = []
                    setattr(self, name, val)

                if model.rid not in [m.rid for m in getattr(self, name)]:
                    getattr(self, name).append(model)
        return self

    @classmethod
    def load(cls, *args, **kwargs):
        raise DeprecationWarning(
            "This method is now depreciated as of version 0.1.0. Trident now requires"
            " model instantiations to be explicitly attached to an AqSession object."
            "\nPlease use the following"
            " methods to initialize your models, which will automatically attach your "
            "session object to"
            " the newly constructed instance."
            "\n(1) `session.{name}.new(*args, **kwargs)` to initialize a new model "
            "using a constructor."
            "\n(2) `session.{name}.load(data)  # to initialize a model with data."
            "\n(3) `{name}.load_from(data, session)".format(name=cls.__name__)
        )

    @classmethod
    def load_from(
        cls, data: dict, owner: "ModelBase" = None
    ) -> Union[List["ModelBase"], "ModelBase"]:
        """Create a new model instance from loaded attributes.

        'obj' should have a o
        """
        if isinstance(data, list):
            models = []
            for d in data:
                model = cls._set_data(d, owner)
                models.append(model)
            return models
        else:
            model = cls._set_data(data, owner)
        return model

    # TODO: rename reload to something else, implement 'refresh' method and
    #       associated tests
    def reload(self, data: dict) -> "ModelBase":
        """Reload model attributes from new data.

        :param data: data to update model instance
        :type data: dict
        :return: model instance
        :rtype: ModelBase
        """
        data_copy = deepcopy(data)
        temp_model = self.__class__.load_from(data_copy, self.session)
        vars(self).update(vars(temp_model))

        return self

    def refresh(self):
        """Refresh this model from data from the server.

        :return: self
        :rtype: self
        """
        interface = self.create_interface()
        interface._do_load = False
        data = interface.find(self.id)
        self.reload(data)
        return self

    @classmethod
    def get_relationships(cls):
        grouped = cls._model_schema.grouped_fields
        relationships = grouped[fields.Relationship.__name__]
        alias = grouped[fields.Alias.__name__]
        for aname, alias_field in alias.items():
            aliased = relationships.get(alias_field.alias, None)
            if aliased:
                relationships[aname] = aliased
        return relationships

    @property
    def session(self):
        """The connected session instance."""
        return self._session

    @session.setter
    def session(self, new_session: "AqSession"):
        """Sets the session of the model.

        .. versionchanged:: 0.1.5a8
            SessionAlreadySet exception will be raised (instead of Exception)
            if session is already set.

        :param new_session: the new session
        :return: None
        :raises SessionAlreadySet: if session is already set
        :raises NoSessionError: if session is not a SessionABC type
        """
        if new_session is not None and not issubclass(type(new_session), SessionABC):
            raise NoSessionError(
                "Cannot instantiate new model because its data parent "
                "session is a type '{}', not a Session object".format(type(new_session))
            )
        if self.session is not None:
            raise SessionAlreadySet(
                "Cannot set session. Model {} already has a session.".format(self)
            )
        self._session = new_session

    def connect_to_session(self, session: "AqSession"):
        """Connect model to a session.

        :param session: the :class:`AqSession <pydent.aqsession.AqSession>`
        :return: None
        :raises SessionAlreadySet: if session is already set
        :raises NoSessionError: if session is not a SessionABC type
        """
        self.session = session

    def _check_for_session(self):
        """Raises error if model is not connected to a session.

        :raises NoSessionError
        """
        if self.session is None:
            raise NoSessionError(
                "No AqSession instance found for '{name}' but one is required for the "
                "method.\nDo one of the following:"
                "\n(1) - Use 'connect_to_session' after initializing your model."
                "\n(2) - If initializing a model use the session model constructor."
                "\n\t >> USE: \t\t'session.{name}.new(*args, **kwargs)'"
                "\n\t >> DO NOT USE:\t'{name}(*args, **kwargs)'"
                "\n(3) - If initializing with load"
                "\n\t >> USE: \t\t'session.{name}.load(data)"
                "\n\t >> DO NOT USE\t'{name}.load(data)".format(
                    name=self.__class__.__name__
                )
            )

    def no_getter(self, *_):
        """Callback that always returns None."""
        return None

    def create_interface(self) -> Union[QueryInterface, BrowserInterface]:
        return self.interface(self.session)

    @classmethod
    def interface(cls, session: "AqSession") -> Union[QueryInterface, BrowserInterface]:
        """Creates a model interface from this class and a session.

        This method can be overridden in model definitions for special
        cases.
        """
        return session.model_interface(cls.__name__)

    @classmethod
    def find(cls, session: "AqSession", model_id: Union[int, str]) -> "ModelBase":
        """Finds a model instance by its model_id."""
        interface = cls.interface(session)
        return interface.find(model_id)

    @classmethod
    def where(
        cls, session: "AqSession", params: Dict
    ) -> Union[None, List["ModelBase"]]:
        """Finds a list of models by some parameters."""
        if params is None:
            return None
        interface = cls.interface(session)
        return interface.where(params)

    @classmethod
    def one(cls, session: "AqSession", query: Dict, **kwargs) -> "ModelBase":
        interface = cls.interface(session)
        query = dict(query)
        query.update(kwargs)
        return interface.one(query)

    def one_callback(self, model_name: str, *args, **kwargs) -> "ModelBase":
        self._check_for_session()
        model = ModelRegistry.get_model(model_name)
        return model.one(self.session, *args, **kwargs)

    def find_callback(self, model_name: str, model_id: Union[str, int]) -> "ModelBase":
        """Finds a model using the model interface and model_id.

        Used to find models in model relationships.
        """
        self._check_for_session()
        if model_id is None:
            return None
        model = ModelRegistry.get_model(model_name)
        self.session._log_to_aqhttp(
            "CALLBACK '{clsname}(rid={rid})' made a FIND request for '{model}'".format(
                clsname=self.__class__.__name__, rid=self.rid, model=model_name
            )
        )
        return model.find(self.session, model_id)

    def where_callback(
        self, model_name: str, *args, **kwargs
    ) -> Union[None, List["ModelBase"]]:
        """Finds models using a model interface and a set of parameters.

        Used to find models in model relationships.
        """
        self._check_for_session()
        query_arg = args[0]

        if isinstance(query_arg, dict):
            if len(query_arg) == 1 and list(query_arg.values())[0] is None:
                return None
            if None in query_arg.values():
                return None
        model = ModelRegistry.get_model(model_name)
        if kwargs is None:
            kwargs = {}
        self.session._log_to_aqhttp(
            "CALLBACK '{clsname}(rid={rid})' made a WHERE request for '{model}'".format(
                clsname=self.__class__.__name__, rid=self.rid, model=model_name
            )
        )
        return model.where(self.session, query_arg, *args[1:], **kwargs)

    def print(self):
        data = self.dump()
        relationships = self.get_relationships()
        data.update(relationships)
        print(data)

    def __str__(self) -> str:
        return self._to_str("id", "rid")

    def _to_str(self, *attributes, **kwargs) -> str:
        if "rid" not in attributes:
            attributes = list(attributes)
            attributes.append("rid")
        data = self._get_data()
        data.update(kwargs)
        attributes = list(attributes) + list(kwargs)
        return "<{} {}>".format(
            self.__class__.__name__,
            " ".join(["{}={}".format(k, data.get(k, None)) for k in attributes]),
        )

    # TODO: anonymize the keys for relationships as well
    def anonymize(self):
        """Resets the primary key of the model and assigns a new rid.

        Metatypes cannot be annonymized.

        :return: self
        """
        if not self.__class__.__name__.endswith("Type"):
            setattr(self, self.PRIMARY_KEY, None)
            setattr(self, "rid", next(self.counter))
            self.updated_at = None
            self.created_at = None
            self.raw = {}
        return self

    def _anonymize_field_keys(self, keep=None):
        for name, relation in self.get_relationships().items():
            if hasattr(relation, "ref"):
                if relation.nested.endswith("Type"):
                    continue
                elif keep and relation.nested in keep:
                    continue
                else:
                    try:
                        setattr(self, relation.ref, None)
                    except TypeError as e:
                        msg = "An error occured while trying to anonymize relation:\n{}".format(
                            relation
                        )
                        raise e.__class__(msg + "\nerr msg: " + str(e))

    def copy(self, keep: bool = None) -> "ModelBase":
        """Provides a deepcopy of the model, but annonymizes the primary and
        global keys unless class is a metatype (e.g. OperationType, SampleType,
        FieldType) or class name is found in list of 'keep'.

        By default, inventory classes such as Sample, Item, and Collection are 'kept'.

        This specific usecase is that when copying whole plans, that the integrity of
        the inventory used in the operations is maintained. These are the items that
        refer to physical inventory in the laboratory and are referred to by their
        `rids`, and
        so it is important to that any of these inventory are always maintain their
        `rids`.
        Meaning in the lab, their is only ONE instance of the inventory. In Trident
        and Aquarium, there is only ONE instance of inventory that is referred to by
        its rid.

        Similarly, any metatype model must also maintain their `rid`.

        :param keep: list of model classes (as a list of strings) to keep un-anonymous
        :return: copied model
        """
        memo = {}
        copied = deepcopy(self, memo)
        if keep is None:
            keep = self.DEFAULT_COPY_KEEP_UNANONYMOUS

        for m in memo.values():
            if issubclass(type(m), ModelBase):
                if keep is None or m.__class__.__name__ not in keep:
                    m.anonymize()
                m._anonymize_field_keys(keep=keep)
        return copied

    @classmethod
    def _flatten_deserialized_data(cls, models: List["ModelBase"], memo: dict) -> dict:
        """Flattens all of the relationships found in the models, returning a
        rid: model dictionary."""
        if models is None:
            return memo
        for model in models:

            if model is None or model.rid in memo:
                continue
            else:
                memo[model.rid] = model
                data = model._get_deserialized_data()
                for key in model.get_relationships():
                    val = data.get(key, None)
                    if val is None:
                        continue
                    elif isinstance(val, list):
                        cls._flatten_deserialized_data(val, memo)
                    else:
                        cls._flatten_deserialized_data([val], memo)
        return memo

    def _rid_dict(self):
        """Dictionary of all models attached to this model keyed by their
        rid."""
        memo = {}
        self._flatten_deserialized_data([self], memo)
        return memo

    def __copy__(self) -> "ModelBase":
        return self.copy()

    #     return cp
    # def patch(self, json_data):
    #     """
    #     Make a patch request to self using json_data.
    #     Reload model instance with new data.
    #     """
    #     result = self.create_interface().patch(self.id, json_data=json_data)
    #     self.reload(data=result)
    #     return self
    #
    # def patch_with_self(self, **kwargs):
    #     """Update changes to this model instance to Aquarium."""
    #     json_data = self.dump(**kwargs)
    #     return self.patch(json_data=json_data)

    # def __getattribute__(self, name):
    #     """Override getattribute to automatically connect sessions"""
    #     res = super().__getattribute__(name)
    #     if isinstance(res, list) or isinstance(res, SchemaModel):
    #         relationships = object.__getattribute__(
    #             self, "get_relationships")()
    #         if name in relationships:
    #             session = object.__getattribute__(self, 'session')
    #             if isinstance(res, list):
    #                 [m.connect_to_session(session) for m in res]
    #             else:
    #                 res.connect_to_session(session)
    #     return res



"""
Session (:mod:`pydent.aqsession`)
=================================

.. currentmodule:: pydent.aqsession

Session class for interacting with Trident and Aquarium. To initialize
a new session, you'll need your Aquarium login credentials and a url:

.. code-block:: python

    from pydent import AqSession
    session = AqSession('joeshmoe', 'pass123', 'http://myaquariumurl.org`)

.. note::
    To login discretely with a hidden password, you may use
    the :meth:`pydent.login <pydent.login>` method.

After initializing a session, models are accessible from the Session:

.. code-block:: python

    sample = session.Sample.one()
    last50 = session.Item.last(50)
    first10 = session.SampleType.first(10)
    mysamples = session.Sample.last(10, query={'user_id': 66})

.. seealso::
    Models :ref:`Models <models>`
        documentation for more information on how to
        manipulate and query models.
    Temporary Sessions :ref:`Cache <cache>`
        documentation on how to use the session cache to
        speed up queries.


User Objects
------------

The AqSession and Browser are the main interaction object for querying the
Aquarium server. The Browser class provides special methods for speeding
up and caching results from the server, which is covered in
:ref:`Advanced Topics <cache>`.

.. currentmodule: pydent.aqsession

.. autosummary::
    :toctree: generated/

    AqSession
    Browser

.. currentmodule: pydent.planner.planner

The Planner allows users to create new experimental plans and upload them to
the Aquarium server for experiment execution.

.. seealso::
    Planning :ref:`Planning <planning>`


    Advanced Topics :ref:`Advanced Topics <advancedtopics>`


.. autosummary::
    :toctree: generated/

    Planner

Non-User Objects
----------------

These modules provide utility support for Trident. These are not relevant for most
users.

.. currentmodule: pydent.aqsession

.. autosummary::
    :toctree: generated/

    SessionABC
    AqHTTP

Interfaces
^^^^^^^^^^

.. currentmodule:: pydent.interfaces

Interfaces govern how the session searches, loads, and dumps
Aquarium models from the server.

.. autosummary::
    :toctree: generated/

    BrowserInterface
    CRUDInterface
    QueryInterface
    QueryInterfaceABC
    SessionInterface
    UtilityInterface

"""
import inspect
import timeit
import webbrowser
from copy import copy
from decimal import Decimal
from typing import Dict
from typing import Iterable
from typing import List
from typing import Type
from typing import Union

from requests.exceptions import ReadTimeout

from pydent.aqhttp import AqHTTP
from pydent.aql import aql
from pydent.aql import aql_schema
from pydent.base import ModelBase
from pydent.base import ModelRegistry
from pydent.browser import Browser
from pydent.interfaces import BrowserInterface
from pydent.interfaces import QueryInterface
from pydent.interfaces import QueryInterfaceABC
from pydent.interfaces import SessionInterface
from pydent.interfaces import UtilityInterface
from pydent.inventory_updater import save_inventory
from pydent.models import __all__ as allmodels
from pydent.sessionabc import SessionABC


class AqSession(SessionABC):
    """Holds an AqHTTP with login information. Creates SessionInterfaces for
    models.

    .. code-block:: python

        session1 = AqSession(username, password, aquairum_url)
        session1.User.find(1)
        # <User(id=1,...)>
    """

    def __init__(
        self,
        login: str,
        password: str,
        aquarium_url: str,
        name: str = None,
        aqhttp: str = None,
    ):
        """Initializes a new tridennt Session.

        :param login: the Aquarium login for the user
        :type login: str
        :param password: the password for the Aquarium login. This will not be
            stored anywhere. However, login headers for Aquarium will be stored
            in .__aqhttp. This may change in the future with changes
            authentication for the Aquarium API
        :type password: str
        :param aquarium_url: the http formatted Aquarium url
        :type aquarium_url: str
        :param name: (optional) name for this session
        :type name: str or None
        """
        self.name = name  #: optional name of the session
        self._aqhttp = None  #: requests interface
        if login is None and password is None and aquarium_url is None:
            if aqhttp is not None:
                self._aqhttp = aqhttp
            else:
                raise ValueError(
                    "Need either a name, password, and url OR an aqhttp instance."
                )
        else:
            self._aqhttp = AqHTTP(login, password, aquarium_url)
        self._current_user = None
        self._interface_class = QueryInterface
        self._initialize_interfaces()
        self._browser = None  #: the sessions browser
        self._using_cache = False
        self.init_cache()
        self.parent_session = (
            None
        )  #: the parent session, if derived from another session

    @property
    def interface_class(self) -> Type:
        """Returns the session's interface class."""
        return self._interface_class

    @interface_class.setter
    def interface_class(self, c: Type):
        """Sets the session's interface class."""
        if not issubclass(c, QueryInterfaceABC):
            raise ValueError(
                "Interface {} is not a subclass of {}".format(
                    type(c), QueryInterfaceABC
                )
            )
        self._interface_class = c

    def _initialize_interfaces(self):
        """Initializes the session's interfaces."""
        # initialize model interfaces
        for model_name in allmodels:
            self._register_interface(model_name)

    def open(self):
        """Open Aquarium in a web browser window."""
        webbrowser.open(self._aqhttp.url)

    @property
    def session(self) -> "AqSession":
        """Return self."""
        return self

    def set_verbose(self, verbose: bool, tb_limit: int = None):
        self._aqhttp.log.set_verbose(verbose, tb_limit=tb_limit)

    def _log_to_aqhttp(self, msg: str):
        """Sends a log message to the aqhttp's logger."""
        self._aqhttp.log.info(msg)

    def _register_interface(self, model_name: str):
        # get model interface from model class
        model_interface = self.interface_class(model_name, self._aqhttp, self)

        # set interface to session attribute (e.g. session.Sample calls Sample model interface)
        setattr(self, model_name, model_interface)

    def set_timeout(self, timeout_in_seconds: int):
        """Sets the request timeout."""
        self._aqhttp.timeout = timeout_in_seconds

    @property
    def url(self):
        """Returns the aquarium_url for this session."""
        return self._aqhttp.aquarium_url

    @property
    def login(self) -> str:
        """Logs into aquarium, generating the necessary headers to perform
        requests to Aquarium."""
        return self._aqhttp.login

    @property
    def current_user(self) -> "ModelBase":
        """Returns the current User associated with this session.

        Returns None if no user is found (as in cases where the Aquarium
        connection is down).
        """
        if self._current_user is None:
            user = self.User.where({"login": self._aqhttp.login})
            if not user:
                return None
            self._current_user = self.User.where({"login": self._aqhttp.login})[0]
        return self._current_user

    def logged_in(self) -> bool:
        """Returns whether the user is logged in. If the session is able to
        return the User model instance using the session's login credentials,
        the user is considered to be logged in.

        :return: whether user is currently logged in
        :rtype: boolean
        """
        if self.current_user is None:
            return False
        return True

    @property
    def models(self) -> List[str]:
        """Returns list of all models available."""
        return list(ModelRegistry.models.keys())

    def model_interface(
        self, model_name: str, interface_class: SessionInterface = None
    ) -> SessionInterface:
        """Returns model interface by name."""
        if interface_class is None:
            interface_class = self.interface_class
        return interface_class(model_name, self._aqhttp, self)

    @property
    def utils(self):
        """Instantiates a utility interface."""
        return UtilityInterface(self._aqhttp, self)

    def _ping_function(self):
        return self.User.one()

    # TODO: put 'ping' in documentation
    def ping(self, num: int = 5) -> Union[None, int]:
        """Performs a number of simple requests (pings) to estimate the speed
        of the server.

        Displays a message about the average time each ping took.
        """
        ping_function = self._ping_function()
        ping_function_source = inspect.getsource(ping_function).strip()
        try:
            secs = timeit.timeit(ping_function, number=num)
            print(
                "{} pings (using the function '{}')".format(num, ping_function_source)
            )
            print("{} seconds per ping".format("%.2E" % Decimal(secs / num)))
            return secs
        except ReadTimeout as e:
            print("Error: {}".format(e))
            print(
                "The Aquarium server ({}) looks like its down. The function '{}' "
                "raised a {} exception, but should not have.".format(
                    self.url, ping_function_source, ReadTimeout
                )
            )
            return None

    @property
    def browser(self):
        return self._browser

    def init_cache(self):
        self._browser = Browser(self)

    def clear_cache(self):
        self.browser.clear()

    @property
    def using_requests(self) -> bool:
        return self._aqhttp._using_requests

    @using_requests.setter
    def using_requests(self, b: bool):
        if b:
            self._aqhttp.on()
        else:
            self._aqhttp.off()

    @property
    def using_cache(self) -> bool:
        return self._using_cache

    @using_cache.setter
    def using_cache(self, b: bool):
        if b:
            self.interface_class = BrowserInterface
            if self.browser is None:
                self.init_cache()
            self._initialize_interfaces()
            self._using_cache = True
        else:
            self._using_cache = False
            self.interface_class = QueryInterface
            self._initialize_interfaces()

    def copy(self):
        instance = self.__class__(
            None, None, None, self.name, aqhttp=copy(self._aqhttp)
        )
        instance.using_requests = self.using_requests
        instance.using_cache = self.using_cache
        return instance

    def with_cache(
        self,
        using_requests: bool = None,
        using_models: bool = False,
        timeout: int = None,
        verbose=None,
    ) -> "AqSession":
        """

        :param using_requests: if False, ForbiddenRequest will be raised if requests are
        made using the session.
        :param using_models: if True (default: False), derived session will inherit
        the current sessions model_cache
        :param timeout: the requests timeout in seconds
        :param verbose: if True, verbose mode will be activated for the derived session
        :return:
        """
        return self(
            using_cache=True,
            using_models=using_models,
            using_requests=using_requests,
            timeout=timeout,
            using_verbose=verbose,
        )

    def with_requests_off(
        self,
        using_cache: bool = None,
        using_models: bool = True,
        timeout: int = None,
        verbose=None,
    ) -> "AqSession":
        return self(
            using_cache=using_cache,
            using_models=using_models,
            using_requests=False,
            timeout=timeout,
            using_verbose=verbose,
        )

    @staticmethod
    def _swap_sessions(from_session, to_session):
        """Moves models from one session to another."""
        if to_session:
            models = from_session.browser.models
            if to_session.browser:
                to_session.browser.update_cache(models)
            for m in models:
                m._session = to_session

    @classmethod
    def query_schema(cls) -> Dict:
        return aql_schema

    def query(self, data: Dict, use_cache: bool = False) -> List[ModelBase]:
        """Perform a complex query a complex JSON query object.

        Check out the :ref:`JSON Schema page <json_schema>` for more information.

        You may also checkout `AqSession.query_schema` to view the JSON schema
        that validates the input.

        ..code-block:: python

            {
              "__model__": "Item",
              "__description__": "Get available primers in last 7 days",
              "query": {
                "object_type": {
                  "query": {
                    "sample_type": {
                      "query": {
                        "name": "Primer"
                      }
                    }
                  }
                },
                "created_at": {
                  "__time__": {
                    "__days__": -7
                  }
                },
                "location": {
                    "__not__": "deleted"
                }
                "__options__": {
                  "limit": 1
                }
              }
            }

        ..versionadded:: 0.1.5a16
            Added query method for complex queries

        :param data: data query
        :param use_cache: whether to inherit the cache from the provided session
            (default: False)
        :return: list of models fitting the query
        :raises: AquariumQueryLanguageValidationError if input data is invalid.
        """
        return aql(self, data, use_cache=use_cache)

    def __getattr__(self, item):
        if item not in self.__dict__:
            non_private = [k for k in QueryInterface.__dict__ if not k.startswith("_")]
            if item in non_private:
                raise AttributeError(
                    "{s} has no attribute '{item}'.\nDid you mean "
                    "'session.[ModelName].{item}'?".format(s=self.__class__, item=item)
                )
            else:
                msg = ModelRegistry.did_you_mean_model(item, fallback=False)
                if msg:
                    raise AttributeError(
                        "{s} has no attribute '{item}'.\n{msg}".format(
                            s=self.__class__, item=item, msg=msg
                        )
                    )

        return self.__getattribute__(item)

    def __call__(
        self,
        using_cache: bool = None,
        using_requests: bool = None,
        timeout: int = None,
        using_models: bool = None,
        using_verbose: bool = None,
        session_swap: bool = False,
    ) -> "AqSession":
        """Factory call for producing a new Session instance.

        .. versionchanges:: 0.1.5a7
            'session_swap` parameter added.

        :param using_cache:
        :param using_requests:
        :param timeout:
        :param using_models:
        :param using_verbose:
        :param session_swap:
        :return:
        """
        new_session = self.copy()
        new_session.parent_session = self
        if using_cache is not None:
            new_session.using_cache = using_cache
        if using_requests is not None:
            new_session.using_requests = using_requests
        if timeout is not None:
            new_session.set_timeout(timeout)
        if session_swap:
            self._swap_sessions(self, new_session)
        elif using_models and self.browser:
            new_session.browser.update_cache(self.browser.models)
        if using_verbose is not None:
            new_session.set_verbose(using_verbose)
        return new_session

    def __enter__(self) -> "AqSession":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self, "parent_session"):
            self._swap_sessions(self, self.parent_session)

    def __repr__(self) -> str:
        return "<{}(name={}, AqHTTP={}), parent={})>".format(
            self.__class__.__name__, self.name, self._aqhttp, id(self.parent_session)
        )



"""Models related to inventory, like Items, Collections, ObjectTypes, and
PartAssociations."""
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

from .data_associations import DataAssociation
from .sample import Sample
from pydent.base import ModelBase
from pydent.marshaller import add_schema
from pydent.models.controller_mixin import ControllerMixin
from pydent.models.crud_mixin import JSONSaveMixin
from pydent.models.crud_mixin import SaveMixin
from pydent.models.data_associations import DataAssociationSaveContext
from pydent.models.data_associations import DataAssociatorMixin
from pydent.relationships import HasMany
from pydent.relationships import HasManyGeneric
from pydent.relationships import HasManyThrough
from pydent.relationships import HasOne
from pydent.relationships import Raw
from pydent.utils.matrix_mapper import IndexType
from pydent.utils.matrix_mapper import MatrixMapping
from pydent.utils.matrix_mapper import MatrixMappingFactory


@add_schema
class ObjectType(SaveMixin, ModelBase):
    """A ObjectType model that represents the type of container an item is."""

    fields = dict(items=HasMany("Item", "ObjectType"), sample_type=HasOne("SampleType"))

    def __str__(self):
        return self._to_str("id", "name")

    def new_item(self, sample: Union[int, Sample]):
        """Create a new item.

        .. versionadded:: 0.1.5a13

        :param sample: the sample id
        :return: the new item
        """
        if isinstance(sample, int):
            sample_id = sample
            sample = self.session.Sample.find(sample_id)
        elif isinstance(sample, Sample):
            sample_id = sample.id
        else:
            raise TypeError(
                "Sample must be either a sample_id (int) or a" " Sample instance."
            )
        return self.session.Item.new(
            sample=sample, sample_id=sample_id, object_type=self, object_type_id=self.id
        )


class ItemLocationMixin(SaveMixin, JSONSaveMixin):
    def move(self, new_location):
        self.session.utils.move_item(self, new_location)

    def store(self):
        self.session.utils.store_item(self)

    def _get_save_json(self):
        data = self.dump()
        if "location" in data:
            del data["location"]
        return data

    def update(self):
        """Updates the item.

        Will update the data associations and its location.
        """
        to_location = self.location
        JSONSaveMixin.save(self, do_reload=True)
        if to_location and to_location != self.location:
            self.move(self.location)
        return self

    def delete(self):
        raise ValueError
        self.move("deleted")

    def mark_as_deleted(self):
        self.location = "deleted"

    def is_deleted(self):
        return self.location == "deleted"

    @abstractmethod
    def create(self):
        pass

    def make(self):
        """Makes the Item on the Aquarium server.

        Requires this Item to be connected to a session.
        """
        if not self.DID_ITEM_WARNING:
            raise DeprecationWarning("This method is depreciated. Use `save()`")
        self.DID_ITEM_WARNING = True
        self.create()


@add_schema
class Item(DataAssociatorMixin, ItemLocationMixin, ModelBase):
    """A physical object in the lab, which a location and unique id."""

    DID_ITEM_WARNING = False

    fields = dict(
        sample=HasOne("Sample"),
        object_type=HasOne("ObjectType"),
        data_associations=HasManyGeneric(
            "DataAssociation", additional_args={"parent_class": "Item"}
        ),
        data=Raw(),
        ignore=("locator_id",),
        part_associations=HasMany(
            "PartAssociation", ref="part_id"
        ),  # TODO: add to change log
        collections=HasManyThrough(
            "Collection", "PartAssociation"
        ),  # TODO: add to change log
    )
    query_hook = {"methods": ["is_part"]}

    def __init__(
        self=None,
        sample_id=None,
        sample=None,
        object_type=None,
        object_type_id=None,
        location=None,
    ):
        if sample_id is None:
            if sample and sample.id:
                sample_id = sample.id

        if object_type_id is None:
            if object_type and object_type.id:
                object_type_id = object_type.id
        super().__init__(
            object_type_id=object_type_id,
            object_type=object_type,
            sample_id=sample_id,
            sample=sample,
            location=location,
        )

    def create(self):
        with DataAssociationSaveContext(self):
            result = self.session.utils.create_items([self])
            self.reload(result[0]["item"])
        return self

    @property
    def containing_collection(self):
        """Returns the collection of which this Item is a part.

        Returns the collection object if the Item is a part, otherwise
        returns None.
        """
        if not self.is_part:
            return None

        assoc_list = self.session.PartAssociation.where({"part_id": self.id})
        if not assoc_list:
            return

        if len(assoc_list) != 1:
            return None

        part_assoc = next(iter(assoc_list))
        if not part_assoc:
            return None

        return self.session.Collection.find(part_assoc.collection_id)

    def as_collection(self):
        """Returns the Collection object with the ID of this Item, which must
        be a collection.

        Returns None if this Item is not a collection.
        """
        if not self.is_collection:
            return None

        return self.session.Collection.find(self.id)

    @property
    def is_collection(self):
        """Returns True if this Item is a collection in a PartAssociation.

        Note: this is not how Aquarium does this test in the `collection?` method.
        """
        assoc_list = self.session.PartAssociation.where({"collection_id": self.id})
        return bool(assoc_list)

    # TODO: add to change log
    @property
    def collection(self):
        return self.collections[0]

    # TODO: add to change log
    @property
    def part_association(self):
        return self.part_associations[0]


@add_schema
class PartAssociation(JSONSaveMixin, ModelBase):
    """Represents a PartAssociation linking a part to a collection.

    Collections contain many `parts`, each of which can refer to a
    different sample.
    """

    fields = dict(part=HasOne("Item", ref="part_id"), collection=HasOne("Collection"))

    def __init__(self, part_id=None, collection_id=None, row=None, column=None):
        super().__init__(
            part_id=part_id, collection_id=collection_id, row=row, column=column
        )

    def get_sample_id(self) -> Union[None, int]:
        if self.sample_id:
            return self.sample_id
        return self.sample.id

    def has_unsaved_sample(self) -> bool:
        if self.is_deserialized("part") and self.part.is_deserialized("sample"):
            if self.part.sample.id is None:
                return True
        return False

    def is_empty(self) -> bool:
        if self.get_sample_id():
            return True
        return False


@add_schema
class Collection(
    ItemLocationMixin, DataAssociatorMixin, SaveMixin, ControllerMixin, ModelBase
):
    """A Collection model, such as a 96-well plate, which contains many
    `parts`, each of which can be associated with a different sample."""

    fields = dict(
        object_type=HasOne("ObjectType"),
        data_associations=HasManyGeneric(
            "DataAssociation", additional_args={"parent_class": "Collection"}
        ),
        part_associations=HasMany("PartAssociation", "Collection"),
        parts=HasManyThrough("Item", "PartAssociation", ref="part_id"),
    )
    query_hook = {"methods": ["dimensions"]}

    # TODO: validate dimensions is returning the same dimensions as that in the object_type
    # TODO: init should establish dimensions

    def __init__(
        self,
        object_type: ObjectType = None,
        location: str = None,
        data_associations: List = None,
        parts: List = None,
        part_associations: List = None,
        **kwargs,
    ):
        """Initialize a new Collection.

        .. versionchanged:: 0.1.5a10
            Advanced indexing added for setting and getting samples and data associations

        **Setting samples using new advanced indexing**

        .. code-block:: python

            object_type = session.ObjectType.one(query='rows > 2 AND columns > 2')
            collection = session.Collection.new(object_type=object_type)

            # assign sample '1' to (0, 0) row=0, column=0
            collection[0, 0] = 1

            # assign sample '2' to (1, 2) row=1, column=2
            collection[1, 2] = 2

            # assign sample '3234' to row 3
            collection[3] = 3234

            # assign sample '444' to column 1
            collection[:, 1] = 444

            # assign sample '6' to the whole collection
            collection[:, :] = 6

            # assign sample using Sample instance
            collection[2, 2] = session.Sample.one()

        **Getting samples using new advanced indexing

        .. code-block:: python

            # get 2d matrix of sample ids
            print(collection.matrix)  # or collection.sample_id_matrix

            # get 2d matrix of Samples assigned at each location
            print(collection.sample_matrix)

            # get 2d matrix of Parts assigned at each location
            print(collection.part_matrix)

            # get 2d matrix of PartAssociations assigned at each location
            collection.part_associations_matrix

            # get 2d matrix of values of DataAssociations at each location
            collection.data_matrix

            # get 2d matrix of DataAssociations at each location
            collection.data_association_matrix

        **Assigning data to locations**

        To assign data, you can use the advanced indexing on the `data_matrix`

        .. code-block:: python

            collection.data_matrix[0, 0] = {'key': 'value'}

            collection.data_matrix[1] = {'key': 'value2'}

            collection.associate_to('key', 'value3', 3, 3)

        You can delete associations using the following:

        .. code-block:: pythong

            # delete 3, 3
            collection.delete_association_at('key', 3, 3)

            # delete first three rows at column 3
            collection.delete_association_at('key', slice(None, 3, None), 3)

            # delete all of the 'key' associations
            collection.delete_association_at('key', slice(None, None, None), slice(None, None, None))


        :param object_type:
        :param location:
        :param data_associations:
        :param parts:
        :param part_associations:
        :param kwargs:
        """
        if isinstance(object_type, ObjectType):
            object_type = object_type
            object_type_id = object_type.id
            dims = object_type.rows, object_type.columns
        else:
            object_type_id = None
            dims = None

        super().__init__(
            object_type=object_type,
            object_type_id=object_type_id,
            location=location,
            dimensions=dims,
            data_associations=data_associations,
            part_associations=part_associations,
            parts=parts,
            **kwargs,
        )

    def _empty(self):
        nrows, ncols = self.dimensions
        data = []
        for r in range(nrows):
            data.append([None] * ncols)
        return data

    def __part_association_matrix(self):
        data = self._empty()
        if self.part_associations:
            for assoc in self.part_associations:
                data[assoc.row][assoc.column] = assoc
        return data

    @staticmethod
    def _get_part(assoc):
        if assoc is not None:
            return assoc.part

    @staticmethod
    def _get_sample_id(assoc):
        if assoc is not None:
            if assoc.part:
                if assoc.part.sample_id:
                    return assoc.part.sample_id
                elif assoc.part.sample:
                    return assoc.part.sample.id
                return None

    @staticmethod
    def _get_sample(assoc):
        if assoc is not None:
            if assoc.part:
                return assoc.part.sample

    @staticmethod
    def _get_data_association(assoc):
        if assoc is not None:
            if assoc.part:
                if assoc.data_associations:
                    return {a.key: a for a in assoc.part.data_associations}
                else:
                    return {}

    @staticmethod
    def _get_data_value(assoc):
        if assoc is not None:
            if assoc.part:
                if assoc.part.data_associations:
                    return {a.key: a.value for a in assoc.part.data_associations}
                else:
                    return {}

    @staticmethod
    def _no_setter(x):
        raise ValueError("Setter is not implemented.")

    def _set_sample(
        self,
        data: List[List[PartAssociation]],
        r: int,
        c: int,
        sample: Union[int, Sample],
    ):
        if data[r][c]:
            part = data[r][c].part
        else:
            part = self.session.Item.new(
                object_type_id=self.session.ObjectType.find_by_name("__Part").id
            )
            association = self.session.PartAssociation.new(
                part_id=part.id, collection_id=self.id, row=r, column=c
            )
            association.part = part
            self.append_to_many("part_associations", association)
        if isinstance(sample, int):
            part.sample_id = sample
            part.reset_field("sample")
        elif isinstance(sample, Sample):
            part.sample = sample
            part.sample_id = sample.id
        elif sample is None:
            part.sample = None
            part.sample_id = None
        else:
            raise ValueError("{} must be a Sample instance or an int".format(sample))

    def _set_key_values(
        self,
        data: List[List[PartAssociation]],
        r: int,
        c: int,
        data_dict: Dict[str, Any],
    ):
        part = None
        if data[r][c]:
            part = data[r][c].part

        if not part:
            raise ValueError(
                "Cannot set data to ({r},{c}) because "
                "there is no Sample assigned to that location".format(r=r, c=c)
            )
        for k, v in data_dict.items():
            part.associate(k, v)

    @property
    def _mapping(self):
        factory = MatrixMappingFactory(self.__part_association_matrix())
        default_setter = self._no_setter
        factory.new("part_association", setter=default_setter, getter=None)
        factory.new("part", setter=default_setter, getter=self._get_part)
        factory.new("sample_id", setter=self._set_sample, getter=self._get_sample_id)
        factory.new("sample", setter=default_setter, getter=self._get_sample)
        factory.new("data", setter=self._set_key_values, getter=self._get_data_value)
        factory.new(
            "data_association", setter=default_setter, getter=self._get_data_association
        )
        return factory

    @property
    def matrix(self):
        """Returns the matrix of Samples for this Collection.

        (Consider using samples of parts directly.)

        .. versionchanged:: 0.1.5a9
            Refactored using MatrixMapping
        """
        return self.sample_id_matrix

    @property
    def part_matrix(self) -> MatrixMapping[Item]:
        """Return a view of :class:`ITem <pydent.models.Item>`

        .. versionadded:: 0.1.5a9

        :return: collection as a view of Items (Parts)
        """
        return self._mapping["part"]

    @property
    def part_association_matrix(self) -> MatrixMapping[PartAssociation]:
        """Return a view of :class:`PartAssociation.

        <pydent.models.PartAssociation>`

        .. versionadded:: 0.1.5a9

        :return: collection as a view of PartAssociations
        """
        return self._mapping["part_association"]

    @property
    def sample_id_matrix(self) -> MatrixMapping[int]:
        """Return a view of sample_ids :class:`Sample <pydent.models.Sample>`

        .. versionadded:: 0.1.5a9

        :return: collection as a view of Sample.ids
        """
        return self._mapping["sample_id"]

    @property
    def sample_matrix(self) -> MatrixMapping[Sample]:
        """Return a view of :class:`Sample <pydent.models.Sample>`

        .. versionadded:: 0.1.5a9

        :return: collection as a view of Samples
        """
        return self._mapping["sample"]

    @property
    def data_matrix(self) -> MatrixMapping[Any]:
        """Return a view of values from the :class:`DataAssociation.

        <pydent.models.DataAssociation>`

        .. versionadded:: 0.1.5a9

        :return: collection as a view of DataAssociation values
        """
        return self._mapping["data"]

    @property
    def data_association_matrix(self) -> MatrixMapping[DataAssociation]:
        """Return a view of :class:`DataAssociation.

        <pydent.models.DataAssociation>`

        .. versionadded:: 0.1.5a9

        :return: collection as a view of DataAssociations
        """
        return self._mapping["data_association"]

    def part(self, row, col) -> Item:
        """Returns the part Item at (row, col) of this Collection (zero-
        based)."""
        return self.part_matrix[row, col]

    def as_item(self):
        """Returns the Item object with the ID of this Collection."""
        return self.session.Item.find(self.id)

    # TODO: implement save and create
    def create(self):
        """Create a new empty collection on the server."""
        result = self.session.utils.model_update(
            "collections", self.object_type_id, self.dump()
        )
        self.id = result["id"]
        self.created_at = result["created_at"]
        self.updated_at = result["updated_at"]
        self.update()
        return self

    def _validate_for_update(self):
        if not self.id:
            raise ValueError(
                "Cannot update part associations since the Collection " "is not saved."
            )
        for association in self.part_associations:
            if association.has_unsaved_sample():
                raise ValueError(
                    "Cannot update. Collection contains Samples ({r},{c})"
                    " that have not yet been saved.".format(
                        r=association.row, c=association.column
                    )
                )

    def update(self):
        self._validate_for_update()
        self.move(self.location)
        for association in self.part_associations:
            if not association.collection_id:
                association.collection_id = self.id
            association.part.save()
            association.part_id = association.part.id
            association.save()
        self.refresh()

    def assign_sample(self, sample_id: int, pairs: List[Tuple[int, int]]):
        """Assign sample id to the (row, column) pairs for the collection.

        :param sample_id: the sample id to assign
        :param pairs: list of (row, column) tuples
        :return: self
        """
        self.controller_method(
            "assign_sample",
            self.get_tableized_name(),
            self.id,
            data={"sample_id": sample_id, "pairs": pairs},
        )
        self.refresh()
        return self

    def remove_sample(self, pairs: List[Tuple[int, int]]):
        """Clear the sample_id assigment in the (row, column) pairs for the
        collection.

        :param pairs: list of (row, column) tuples
        :return: self
        """
        self.controller_method(
            "delete_selection",
            self.get_tableized_name(),
            self.id,
            data={"pairs": pairs},
        )
        self.refresh()
        return self

    def associate_to(self, key, value, r: int, c: int):
        self.data_matrix[r, c] = {key: value}

    def delete_association_at(self, key, r: int, c: int):
        part_matrix = self.part_matrix
        for r, c in part_matrix._iter_indices((r, c)):
            part = part_matrix[r, c]
            part.delete_data_associations(key)

    def __getitem__(self, index: IndexType) -> Union[int, List[int], List[List[int]]]:
        """Returns the sample_id of the part at the provided index.

        :param index: either an int (for rows), tuple of ints/slice objs, (row, column),
            or a slice object.
        :return: sample_ids
        """
        return self.matrix[index]

    def __setitem__(self, index: IndexType, sample: Union[int, Sample]):
        """Sets the sample_ids of the collection, creating the PartAssociations
        and parts if needed.

        :param index: either an int (for rows), tuple of ints/slice objs, (row, column),
            or a slice object.
        :param sample: either a Sample with a valid id or sample_id
        :return:
        """
        self.sample_id_matrix[index] = sample

    # TODO: add data associations to matrix
    # TODO: collections test
    # def update(self):
    #     self.as_item().update()
    #
    # def save(self):
    #     self.as_item().save()



from typing import Union


class ControllerMixin:
    def controller_method(
        self,
        controller_method: str,
        table: str,
        model_id: Union[str, int, None],
        data: Union[None, dict],
        params=None,
    ) -> dict:
        """Method for create, updating, and deleting models.

        :param controller_method: Name of the controller method
        :param table: Table name of model (e.g. 'samples' or 'data_associations')
        :param model_id: Optional model_id (not required for 'post')
        :param data: data
        :param params
        :return: json formatted server response
        :rtype: dict
        """

        if model_id:
            url = "{}/{}/{}".format(table, model_id, controller_method)
        else:
            url = "{}/{}".format(table, controller_method)

        result = self.aqhttp.request("post", url, json=data, params=params)
        return result



"""Utilities

==========
utils
==========

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    async_requests
    logger

"""
import pprint as pprint_module

from .loggable import condense_long_lists
from .loggable import Loggable
from .loggable import pprint_data
from .query_builder import QueryBuilder
from pydent.utils.async_requests import make_async

printer = pprint_module.PrettyPrinter(indent=1)
pprint = printer.pprint
pformat = printer.pformat


def filter_list(objlist, **kwargs):
    """Filters a list of objects based on attributes in kwargs."""
    intersection = []
    for obj in objlist:
        is_ok = True
        for k in kwargs:
            if not hasattr(obj, k):
                is_ok = False
                break
            if not getattr(obj, k) == kwargs[k]:
                is_ok = False
                break
        if is_ok:
            intersection.append(obj)
    return type(objlist)(intersection)


def url_build(*parts):
    """Join parts of a url into a string."""
    url = "/".join(p.strip("/") for p in parts)
    return url


def empty_copy(obj):
    """Return an empty copy of an object for copying purposes."""

    class Empty(obj.__class__):
        def __init__(self):
            pass

    newcopy = Empty()
    newcopy.__class__ = obj.__class__
    return newcopy


logger = Loggable("pydent")
# def filter_dictionary(dictionary, filter_function):
#     new_dict = {}
#     if isinstance(dictionary, dict):
#         for key, val in dictionary.items():
#             if filter_function(key, val):
#                 new_dict[key] = val
#             if isinstance(val, dict):
#                 new_dict[key] = filter_dictionary(val, filter_function)
#         return new_dict

# def ignore_none(d):
#     return filter_dictionary(d, lambda k, v: v is not None)
#
# def ignore_empty(d):
#     return filter_dictionary(d, lambda k, v: v != [])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydent',
 'pydent.aql',
 'pydent.marshaller',
 'pydent.models',
 'pydent.planner',
 'pydent.utils',
 'pydent.utils.loggable']

package_data = \
{'': ['*'],
 'pydent': ['.idea/*', '.idea/inspectionProfiles/*', '.idea/libraries/*']}

install_requires = \
['colorlog>=4.0,<5.0',
 'inflection>=0.3.1,<0.4.0',
 'nest-asyncio>=1.0,<2.0',
 'networkx>=2.3,<3.0',
 'requests>=2.22,<3.0',
 'retry>=0.9.2,<0.10.0',
 'tqdm>=4.32,<5.0']

setup_kwargs = {
    'name': 'pydent',
    'version': '0.1.5a18',
    'description': "Aquarium's Python API for planning, executing, and analyzing scientific experiments.",
    'long_description': '# Trident: Aquarium API Scripting\n\n[![CircleCI](https://circleci.com/gh/klavinslab/trident/tree/master.svg?style=svg&circle-token=88677c59698d55a127a080cba9ca025cf8072f6c)](https://circleci.com/gh/klavinslab/trident/tree/master)\n[![PyPI version](https://badge.fury.io/py/pydent.svg)](https://badge.fury.io/py/pydent)\n\nTrident is the python API scripting for Aquarium.\n\n## Documentation\n\n[API documentation can be found here at klavinslab.org/trident](http://www.klavinslab.org/trident)\n\n## Requirements\n\n* Python > 3.4\n* An Aquarium login\n\n## Quick installation\n\nPydent can be installed using `pip3`.\n\n```\n    pip3 install pydent\n```\n\nor upgraded using\n\n```\n    pip3 install pydent --upgrade\n```\n\n## Basic Usage\n\n### Logging in\n\n```python\nfrom pydent import AqSession\n\n# open a session\nmysession = AqSession("username", "password", "www.aquarium_nursery.url")\n\n# find a user\nu = mysession.User.find(1)\n\n# print the user data\nprint(u)\n```\n\n### Models\n\n```python\nprint(mysession.models)\n```\n\n#### Finding models\n\n* By name: `nursery.SampleType.find_by_name("Primer")`\n\n* By ID: `nursery.SampleType.find(1)`\n\n* By property: `nursery.SampleType.where({\'name\': \'Primer\'})`\n\n* All models: `nursery.SampleType.all()`\n\n#### Getting nested data\n\n```python\n# samples are linked to sample_type\nprimer_type = mysession.SampleType.find_by_name("Primer")\nprimers = primer_type.samples\n\n# and sample type is linked to sample\np = primers[0]\nprint(p.sample_type)\n```\n\n#### Available nested relationships\n\n```python\nprimer_type = mysession.SampleType.find(1)\nprint(primer_type.relationships)\n```',
    'author': 'jvrana',
    'author_email': 'justin.vrana@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.github.com/klavinslab/trident',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from abc import ABC


class SessionABC(ABC):
    """Session abstract base class."""

    Account = None  #: Account model interface
    AllowableFieldType = None  #: AllowableFieldType model interface
    Budget = None  #: Budget model interface
    Code = None  #: Code model interface
    Collection = None  #: Collection model interface
    DataAssociation = None  #: DataAssociation model interface
    FieldType = None  #: FieldType model interface
    FieldValue = None  #: FieldValue model interface
    Group = None  #: Group model interface
    Invoice = None  #: Invoice model interface
    Item = None  #: Item model interface
    Job = None  #: Job model interface
    JobAssociation = None  #: JobAssociation model interface
    Library = None  #: Library model interface
    Membership = None  #: Membership model interface
    ObjectType = None  #: ObjectType model interface
    Operation = None  #: Operation model interface
    OperationType = None  #: OperationType model interface
    PartAssociation = None  #: PartAssociation model interface
    Plan = None  #: Plan model interface
    PlanAssociation = None  #: PlanAssociation model interface
    Sample = None  #: Sample model interface
    SampleType = None  #: SampleType model interface
    Upload = None  #: Upload model interface
    User = None  #: User model interface
    UserBudgetAssociation = None  #: UserBudgetAssociation model interface
    Wire = None  #: Wire model interface
